package org.openxma.dsl.common.formatter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.Action;
import org.eclipse.xtext.IGrammarAccess;
import org.eclipse.xtext.TerminalRule;
import org.eclipse.xtext.formatting.impl.BaseFormatter;
import org.eclipse.xtext.formatting.impl.BaseTokenStream;
import org.eclipse.xtext.parsetree.AbstractNode;
import org.eclipse.xtext.parsetree.NodeAdapter;
import org.eclipse.xtext.parsetree.NodeUtil;
import org.eclipse.xtext.parsetree.reconstr.ITokenStream;
import org.openxma.dsl.common.formatter.locator.LineWrapLocator;
import org.openxma.dsl.common.formatter.locator.SimpleLocator;

abstract public class OpenXMAFormatter extends BaseFormatter {

	public class CustomOneWhitespaceFormatterStream extends BaseTokenStream {

		protected boolean hasStarted;

		protected boolean preserveWS;

		protected boolean wsWritten;
		
		protected boolean newLineWritten;
		
		protected boolean allowSpace;
		
		protected EObject previousGrammarElement;
		
		OpenXMAFormattingConfig formattingConfig;		
		
		int indentionCount;		

		public CustomOneWhitespaceFormatterStream(ITokenStream out, boolean preserve) {
			super(out);
			this.hasStarted = false;
			this.wsWritten = false;
			this.newLineWritten = false;
			this.indentionCount = 0;	
			this.allowSpace = true;
			this.previousGrammarElement = null;			
			this.preserveWS = preserve;
			formattingConfig = new OpenXMAFormattingConfig();			
			configureFormatting(formattingConfig);
		}

		protected boolean isComment(EObject grammarElement, String value) {
			return grammarElement != getWSRule();
		}

		public void writeHidden(EObject grammarElement, String value) throws IOException {
			NodeAdapter nodeAdapter = NodeUtil.getNodeAdapter(grammarElement);
			initGrammarElement(nodeAdapter!=null?nodeAdapter.getParserNode():null,grammarElement);			
			boolean isCmt = isComment(grammarElement, value);
			/* Format multi line comment */
			if (isCmt && grammarElement instanceof TerminalRule && "ML_COMMENT".equals(((TerminalRule)grammarElement).getName())) {
				value = formatCommentLines(value);				
			}
			if (preserveWS || isCmt) {
				if (isCmt) {
					writeSpaceIfNecessary(grammarElement);
					hasStarted = true;			
				}
				out.writeHidden(grammarElement, value);
				if (isCmt) {
					finishNewLineWritten();			
				}
				wsWritten = !isCmt;
			}
		}

		/**
		 * Format the lines of a multi line comment:
		 * 
		 * Handle indents and line wraps.
		 * 
		 * @param value
		 * @return
		 */
		private String formatCommentLines(String value) {
			StringBuilder multiLineCommentBuffer = new StringBuilder();
			StringTokenizer tokenizer = new StringTokenizer(value,"\n");
			boolean firstLine = true;
			while (tokenizer.hasMoreTokens()) {
				String commentLine = tokenizer.nextToken();
				/* add indention */
				if (!firstLine) {
					int firstContentCharIndex =-1;
					for (int i=0;i<commentLine.length();i++) {
						char c = commentLine.charAt(i);
						if (c != ' ' && c != '\t') {
							firstContentCharIndex = i;
							break;
						}
					}
					if (firstContentCharIndex >= 0) {
						String spacesBeforeStar = formattingConfig.getIndention(indentionCount)+" ";
						String textOfComment = commentLine.substring(firstContentCharIndex);
						if (textOfComment.length()>0 && textOfComment.charAt(0) != '*') {
							spacesBeforeStar+="  ";
						}
						commentLine = spacesBeforeStar+textOfComment;
					}
					commentLine = "\n"+commentLine;
				}
				firstLine=false;
				multiLineCommentBuffer.append(commentLine);
			}				
			value = multiLineCommentBuffer.toString();
			/* add line wrap at the end*/
			value+="\n";
			return value;
		}

		public void writeSemantic( EObject grammarElement, String value)
				throws IOException {
			NodeAdapter nodeAdapter = NodeUtil.getNodeAdapter(grammarElement);
			initGrammarElement(nodeAdapter!=null?nodeAdapter.getParserNode():null,grammarElement);			
			writeSpaceIfNecessary(grammarElement);
			hasStarted = true;
			out.writeSemantic(grammarElement, value);
			wsWritten = false;
			allowSpace = true;
			newLineWritten = false;				
		}

		protected void writeSpaceIfNecessary(EObject grammarElement) throws IOException {
			if (allowSpace && hasStarted && !wsWritten && !(grammarElement instanceof Action)) {
				if (newLineWritten == false) {
					out.writeHidden(getWSRule(), " ");
				}			
				wsWritten = true;
			}
		}
		
		public void updateSpaceInformation(AbstractNode node, EObject elementOfLeftNode, EObject elementOfRightNode) {
			if (allowSpace) {
				allowSpace = formattingConfig.allowSpace(node,elementOfLeftNode,elementOfRightNode);
			}
		}

		public void updateIndentionInformation(AbstractNode node, EObject previousGrammarElement, EObject grammarElement) throws IOException {		
			ArrayList<SimpleLocator> startLocators = formattingConfig.startIndentionLocators.findAllLocators(node,previousGrammarElement,grammarElement);
			for (int i=0;i<startLocators.size();i++) {
				indentionCount++;
			}
			ArrayList<SimpleLocator> endLocators = formattingConfig.endIndentionLocators.findAllLocators(node,previousGrammarElement,grammarElement);	
			for (int i=0;i<endLocators.size();i++) {
				indentionCount--;
			}			
		}		
		
		protected void writeNewLine(AbstractNode node, EObject previousGrammarElement, EObject grammarElement) throws IOException {			
			LineWrapLocator lineWrapLocator = (LineWrapLocator)formattingConfig.lineWrapLocators.findFirstLocator(node,previousGrammarElement,grammarElement);
			if (lineWrapLocator != null) {
				IFormatCondition condition = lineWrapLocator.getFormatCondition();
				if (condition == null || condition.applyFormat(node)) {
					for (int i=0;i<lineWrapLocator.getLineCount();i++) {
						out.writeHidden(getWSRule(), "\n");
					}
					finishNewLineWritten();					
				}
			}
		}
		
		public void finishNewLineWritten() throws IOException {
			if (indentionCount > 0) {
				out.writeHidden(getWSRule(),formattingConfig.getIndention(indentionCount));
			}					
			newLineWritten = true;			
		}

		public void initGrammarElement(AbstractNode node, EObject grammarElement) throws IOException {
			updateSpaceInformation(node,previousGrammarElement,grammarElement);
			updateIndentionInformation(node,previousGrammarElement, grammarElement);
			writeNewLine(node,previousGrammarElement,grammarElement);			
			previousGrammarElement = grammarElement; 			
		}		
	
	}
	
	protected IGrammarAccess getGrammarAccess() {
		return grammar;
	}	
	
	abstract public void configureFormatting(OpenXMAFormattingConfig c);	

	public ITokenStream createFormatterStream(String initalIndentation,
			ITokenStream outputStream, boolean preserveWhitespaces) {
		return new CustomOneWhitespaceFormatterStream(outputStream,
				preserveWhitespaces);
	}

}
